mod icon;

use std::{collections::HashMap, fs, io::ErrorKind, path::PathBuf, str::FromStr};

use icon::extract_icon;
use mlua::{
    FromLua, Lua, Number, Table,
    Value::{self as LuaValue},
};
use once_cell::sync::{Lazy, OnceCell};

use axum::{
    extract::Request,
    handler::HandlerWithoutStateExt,
    http::{HeaderName, HeaderValue, StatusCode},
    response::{Html, IntoResponse},
    routing::get,
};
use base64::prelude::*;
use reqwest::Method;
use serde::Serialize;
use tera::{to_value, Context, Result as TeraResult, Tera, Value};
use time::OffsetDateTime;
use tokio::time::Instant;
use tokio_stream::{wrappers::ReadDirStream, StreamExt};
use tower_http::services::ServeDir;

pub type AError = Box<dyn std::error::Error>;
pub type AResult<T> = Result<T, AError>;

pub static CWD: Lazy<PathBuf> = Lazy::new(|| std::env::current_dir().unwrap());

#[derive(Debug, Clone)]
struct ArrayBuffer(Vec<u8>);

impl mlua::FromLua for ArrayBuffer {
    fn from_lua(value: LuaValue, _: &Lua) -> mlua::Result<Self> {
        if let Some(data) = value.as_str() {
            return Ok(Self(data.as_bytes().to_owned()));
        }

        if let Some(data) = value.as_table() {
            let mut out = vec![];
            for pair in data.pairs::<mlua::Value, mlua::Value>() {
                let (_, v) = pair?;
                if let Some(value) = v.as_integer() {
                    out.push(value as u8);
                }
            }
            return Ok(Self(out));
        }

        Err(mlua::Error::FromLuaConversionError {
            from: value.type_name(),
            to: "ArrayBuffer".to_string(),
            message: None,
        })
    }
}

pub static TERA: Lazy<Tera> = Lazy::new(|| {
    let mut tera = Tera::default();
    tera.add_raw_template(
        "index.html.jinja",
        include_str!("../templates/index.html.jinja"),
    )
    .unwrap();
    tera.register_filter("iconize", iconize);
    tera.register_filter("from_ico", from_ico);
    tera.register_filter("size", size);
    tera.register_filter("time", time);
    tera.register_filter("md", md);

    tera
});

pub static ICONS: Lazy<HashMap<&str, String>> = Lazy::new(|| {
    let mut map = HashMap::new();
    map.insert(
        "unknown",
        BASE64_STANDARD.encode(include_bytes!("../icons/unknown.png")),
    );
    map.insert(
        "back",
        BASE64_STANDARD.encode(include_bytes!("../icons/back.png")),
    );
    map.insert(
        "dir",
        BASE64_STANDARD.encode(include_bytes!("../icons/dir.png")),
    );
    map.insert(
        "binary",
        BASE64_STANDARD.encode(include_bytes!("../icons/binary.png")),
    );
    map.insert(
        "image",
        BASE64_STANDARD.encode(include_bytes!("../icons/image2.png")),
    );
    map.insert(
        "compressed",
        BASE64_STANDARD.encode(include_bytes!("../icons/compressed.png")),
    );

    map
});

fn from_ico(value: &Value, _args: &HashMap<String, Value>) -> TeraResult<Value> {
    let Some(icon) = value.as_array() else {
        return iconize(
            &Value::String(String::from("")),
            &std::default::Default::default(),
        );
    };

    let icon = icon
        .iter()
        .map(|n| n.as_u64().unwrap() as u8)
        .collect::<Vec<_>>();

    let data = BASE64_STANDARD.encode(icon);

    let str = format!("<img src=\"data:image/png;base64,{}\" alt=\"\"/>", data);

    Ok(to_value(str).unwrap())
}

fn iconize(value: &Value, _args: &HashMap<String, Value>) -> TeraResult<Value> {
    let icon = value.as_str().unwrap_or("unknown");

    let icon = match icon {
        "DIR" | "BACK" => "dir",
        "image/png" | "image/jpg" | "image/webp" => "image",
        "application/zip" => "compressed",
        "application/octet-stream" => "binary",
        _ => "unknown",
    };

    let icons = &*ICONS;
    let icon = if icons.contains_key(icon) {
        &icons[icon]
    } else {
        println!("Unknown filetype: {}", icon);
        &icons["unknown"]
    };

    let str = format!("<img src=\"data:image/png;base64,{}\" alt=\"\" />", icon);

    Ok(to_value(str).unwrap())
}

fn size(value: &Value, _args: &HashMap<String, Value>) -> TeraResult<Value> {
    let Some(value) = value.as_f64() else {
        return Ok(value.clone());
    };

    Ok(to_value(human_bytes::human_bytes(value)).unwrap())
}

fn time(value: &Value, _args: &HashMap<String, Value>) -> TeraResult<Value> {
    let time_format: std::vec::Vec<time::format_description::FormatItem<'_>> =
        time::format_description::parse("[day]-[month]-[year] [hour]:[minute]:[second]").unwrap();

    let Some(time) = value.as_array() else {
        return Ok(value.clone());
    };

    let time: Vec<i64> = time.iter().map(|v| v.as_i64().unwrap()).collect();

    let new_time: OffsetDateTime = OffsetDateTime::now_utc();
    let new_time = new_time.replace_year(time[0] as i32).unwrap();
    let new_time = new_time.replace_ordinal(time[1] as u16).unwrap();
    let new_time = new_time.replace_hour(time[2] as u8).unwrap();
    let new_time = new_time.replace_minute(time[3] as u8).unwrap();
    let new_time = new_time.replace_second(time[4] as u8).unwrap();
    let new_time = new_time.replace_nanosecond(time[5] as u32).unwrap();

    Ok(to_value(new_time.format(&time_format).unwrap()).unwrap())
}

fn md(value: &Value, _args: &HashMap<String, Value>) -> TeraResult<Value> {
    // parse value as markdown and return html
    if let Some(value) = value.as_str() {
        if let Ok(md) = markdown::to_html_with_options(value, &markdown::Options::gfm()) {
            return Ok(to_value(md).unwrap());
        }
    }

    Ok(to_value(value).unwrap())
}

pub static BASE_DIR: OnceCell<PathBuf> = OnceCell::new();

#[derive(Debug, Serialize)]
struct FileInfo {
    filename: String,
    size: usize,
    format: String,
    icon: Option<Vec<u8>>,
    // created: OffsetDateTime,
    modified: OffsetDateTime,
    description: Option<String>,
}

#[derive(Debug, Clone)]
struct FetchOptions {
    method: String,
    headers: HashMap<String, String>,
    body: Option<ArrayBuffer>,
}

impl Default for FetchOptions {
    fn default() -> Self {
        Self {
            method: "GET".to_string(),
            headers: HashMap::new(),
            body: None,
        }
    }
}

impl FromLua for FetchOptions {
    fn from_lua(value: LuaValue, _lua: &Lua) -> mlua::Result<Self> {
        let Some(options) = value.as_table() else {
            return Err(mlua::Error::FromLuaConversionError {
                from: value.type_name(),
                to: "FetchOptions".to_string(),
                message: Some("options must be a table".to_string()),
            });
        };

        let mut result = FetchOptions::default();
        if let Ok(method) = options.get::<String>("method") {
            result.method = method;
        }
        if let Ok(headers) = options.get::<HashMap<String, String>>("headers") {
            result.headers = headers;
        }
        if let Ok(body) = options.get::<Option<ArrayBuffer>>("body") {
            result.body = body;
        }

        Ok(result)
    }
}

#[tokio::main]
async fn main() -> AResult<()> {
    let listener = tokio::net::TcpListener::bind("[::]:3000").await.unwrap();
    Ok(axum::serve(listener, file_handler.into_make_service()).await?)
}

async fn file_handler(request: Request) -> impl IntoResponse {
    std::env::set_current_dir(&*CWD).unwrap();
    let base_dir: &PathBuf = BASE_DIR.get_or_init(|| {
        PathBuf::from_str(&std::env::var("BASE_DIR").unwrap_or(String::from("./public"))).unwrap()
    });

    let base_dir = base_dir.canonicalize().map_err(|e| {
        dbg!(e);
        (StatusCode::NOT_FOUND, "404: Not Found".to_string())
    })?;

    let mut request_uri_path = request.uri().path().to_owned();

    if request_uri_path.ends_with("/") {
        let mut a = base_dir.clone();
        a.push(".".to_owned() + &request_uri_path);
        a.push("init.lua");
        if let Ok(b) = a.canonicalize() {
            if b.exists() {
                request_uri_path += "init.lua";
            }
        }
    }

    if !request_uri_path.ends_with(".lua") {
        return ServeDir::new(base_dir)
            .fallback(get(handler))
            .try_call(request)
            .await
            .map(|e| e.into_response())
            .map_err(|_| (StatusCode::NOT_FOUND, "404: Not Found".to_string()));
    };

    let lua = Lua::new();
    let globals = lua.globals();

    use mlua::{ExternalResult, LuaSerdeExt};

    let fetch = lua
        .create_function(|lua, params: (String, Option<FetchOptions>)| {
            let uri = params.0;
            let options = params.1;
            let options = options.unwrap_or_default();

            let Ok(method) = Method::from_bytes(options.method.as_bytes()) else {
                return Err(mlua::Error::RuntimeError("Invalid method".to_string()));
            };
            let Ok(uri) = reqwest::Url::parse(&uri) else {
                return Err(mlua::Error::RuntimeError("Invalid uri".to_string()));
            };

            let mut request = reqwest::blocking::Request::new(method, uri);
            // let mut headers = ;

            for (k, v) in &options.headers {
                let Ok(k) = reqwest::header::HeaderName::from_str(k) else {
                    return Err(mlua::Error::RuntimeError("Invalid header name".to_string()));
                };
                let Ok(v) = reqwest::header::HeaderValue::from_str(v) else {
                    return Err(mlua::Error::RuntimeError(
                        "Invalid header value".to_string(),
                    ));
                };

                let _ = request.headers_mut().try_append(k, v);
            }

            *request.body_mut() = options.body.map(|f| f.0.into());

            let client = reqwest::blocking::Client::new();
            let resp = client
                .execute(request)
                .and_then(|resp| resp.error_for_status())
                .into_lua_err()?;

            let json = resp.bytes().into_lua_err()?;
            lua.to_value(&json.to_vec())
        })
        .map_err(|e| {
            eprintln!("Lua Error: {:?}", e);
            render_lua_error(e)
        })?;

    let base64_encode = lua
        .create_function(|lua, bytes: ArrayBuffer| {
            let b64 = BASE64_STANDARD.encode(&bytes.0);
            lua.to_value(&b64)
        })
        .map_err(|e| {
            eprintln!("Lua Error: {:?}", e);
            render_lua_error(e)
        })?;

    let base64_decode = lua
        .create_function(|lua, data: ArrayBuffer| {
            let b64 = BASE64_STANDARD.decode(&data.0).into_lua_err()?;
            lua.to_value(&b64)
        })
        .map_err(|e| {
            eprintln!("Lua Error: {:?}", e);
            render_lua_error(e)
        })?;

    let json_decode = lua
        .create_function(|lua, data: ArrayBuffer| {
            let data = String::from_utf8_lossy(&data.0);
            let json = serde_json::from_str::<serde_json::Value>(&data).into_lua_err()?; //data.json::<serde_json::Value>().into_lua_err()?;
            lua.to_value(&json)
        })
        .map_err(|e| {
            eprintln!("Lua Error: {:?}", e);
            render_lua_error(e)
        })?;

    let to_string = lua
        .create_function(|lua, data: ArrayBuffer| {
            let data = String::from_utf8_lossy(&data.0);
            lua.to_value(&data)
        })
        .map_err(|e| {
            eprintln!("Lua Error: {:?}", e);
            render_lua_error(e)
        })?;

    let dbg = lua
        .create_function(|_, value: LuaValue| {
            dbg!(value);
            Ok(())
        })
        .map_err(|e| {
            eprintln!("Lua Error: {:?}", e);
            render_lua_error(e)
        })?;

    globals.set("fetch", fetch).map_err(|e| {
        eprintln!("Lua Error: {:?}", e);
        render_lua_error(e)
    })?;

    globals.set("base64_encode", base64_encode).map_err(|e| {
        eprintln!("Lua Error: {:?}", e);
        render_lua_error(e)
    })?;

    globals.set("base64_decode", base64_decode).map_err(|e| {
        eprintln!("Lua Error: {:?}", e);
        render_lua_error(e)
    })?;

    globals.set("json_decode", json_decode).map_err(|e| {
        eprintln!("Lua Error: {:?}", e);
        render_lua_error(e)
    })?;

    // globals.set("fetch_b64", fetch_b64).map_err(|e| {
    //     eprintln!("Lua Error: {:?}", e);
    //     render_lua_error(e)
    // })?;

    globals.set("dbg", dbg).map_err(|e| {
        eprintln!("Lua Error: {:?}", e);
        render_lua_error(e)
    })?;

    globals.set("to_string", to_string).map_err(|e| {
        eprintln!("Lua Error: {:?}", e);
        render_lua_error(e)
    })?;

    let request_table = lua.create_table().map_err(|e| {
        eprintln!("Lua Error: {:?}", e);
        render_lua_error(e)
    })?;

    let headers_table = lua
        .create_table_from(
            request
                .headers()
                .into_iter()
                .map(|(k, v)| (k.to_string(), v.to_str().unwrap())),
        )
        .map_err(|e| {
            eprintln!("Lua Error: {:?}", e);
            render_lua_error(e)
        })?;

    let query = request.uri().query().unwrap_or("");

    let query: HashMap<String, String> = serde_qs::from_str(query)
        .map_err(|e| {
            println!("{:#?}", e);
            (
                StatusCode::INTERNAL_SERVER_ERROR,
                "500: Unable to parse query string".to_string(),
            )
        })
        .unwrap_or(HashMap::new());

    let query_table = lua.create_table_from(query).map_err(|e| {
        eprintln!("Lua Error: {:?}", e);
        render_lua_error(e)
    })?;

    request_table.set("headers", headers_table).map_err(|e| {
        eprintln!("Lua Error: {:?}", e);
        render_lua_error(e)
    })?;

    request_table
        .set("uri", request.uri().to_string())
        .map_err(|e| {
            eprintln!("Lua Error: {:?}", e);
            render_lua_error(e)
        })?;

    request_table
        .set("method", request.method().to_string())
        .map_err(|e| {
            eprintln!("Lua Error: {:?}", e);
            render_lua_error(e)
        })?;

    request_table.set("query", query_table).map_err(|e| {
        eprintln!("Lua Error: {:?}", e);
        render_lua_error(e)
    })?;

    // Inject functions to change headers and such
    globals.set("request", request_table).map_err(|e| {
        eprintln!("Lua Error: {:?}", e);
        render_lua_error(e)
    })?;

    let response_table = lua.create_table().map_err(|e| {
        eprintln!("Lua Error: {:?}", e);
        render_lua_error(e)
    })?;

    response_table
        .set(
            "headers",
            lua.create_table().map_err(|e| {
                eprintln!("Lua Error: {:?}", e);
                render_lua_error(e)
            })?,
        )
        .map_err(|e| {
            eprintln!("Lua Error: {:?}", e);
            render_lua_error(e)
        })?;

    globals.set("response", response_table).map_err(|e| {
        eprintln!("Lua Error: {:?}", e);
        render_lua_error(e)
    })?;

    let mut path = base_dir.clone();

    let uri = urlencoding::decode(&request_uri_path[1..]).map_err(|e| {
        println!("{:?}", e);
        (StatusCode::INTERNAL_SERVER_ERROR, e.to_string())
    })?;

    path.push(&*uri);
    let full_path = fs::canonicalize(&path).map_err(|e| {
        println!("{:?}", e);
        (StatusCode::NOT_FOUND, "404: Not Found".to_string())
    })?;

    let full_base_path = fs::canonicalize(base_dir).map_err(|e| {
        println!("{:?}", e);
        (StatusCode::NOT_FOUND, "404: Not Found".to_string())
    })?;

    if !full_path.starts_with(&full_base_path) {
        return Err((StatusCode::BAD_REQUEST, "400: Bad Request".to_string()));
    }

    let script = fs::read_to_string(full_path).map_err(|e| {
        eprintln!("Lua Read Error: {:?}", e);
        (StatusCode::INTERNAL_SERVER_ERROR, e.to_string())
    })?;

    // ------------------------------------------------------------------------

    // TODO: FIX THIS, THIS IS DISGUSTING
    std::env::set_current_dir(full_base_path).unwrap();

    let script = lua.load(script).eval::<LuaValue>().map_err(|e| {
        std::env::set_current_dir(&*CWD).unwrap();
        eprintln!("Lua Error: {:?}", e);
        render_lua_error(e)
    })?;
    std::env::set_current_dir(&*CWD).unwrap();

    let result = lua.load("return response").eval::<Table>().map_err(|e| {
        eprintln!("Lua Error: {:?}", e);
        render_lua_error(e)
    })?;

    // ------------------------------------------------------------------------

    let script = if let LuaValue::String(script) = script {
        script.to_string_lossy().to_string()
    } else {
        serde_json::to_string(&script).map_err(|e| {
            eprintln!("Serde Error: {:?}", e);
            (
                StatusCode::INTERNAL_SERVER_ERROR,
                format!("Serde error: {}", e),
            )
        })?
    };

    let mut response = Html(script).into_response();

    if let Ok(status_code) = result.get::<Number>("code") {
        if let Ok(code) = StatusCode::from_u16(status_code as u16) {
            *response.status_mut() = code;
        }
    }

    if let Ok(headers) = result.get::<Table>("headers") {
        let pairs = headers.pairs::<String, String>();

        for pair in pairs {
            let (k, v) = pair.map_err(|e| {
                eprintln!("Lua Error: {:?}", e);
                render_lua_error(e)
            })?;
            response.headers_mut().insert(
                HeaderName::from_str(&k).map_err(|e| {
                    eprintln!("Response Error: {:?}", e);
                    (StatusCode::INTERNAL_SERVER_ERROR, e.to_string())
                })?,
                HeaderValue::from_str(&v).map_err(|e| {
                    eprintln!("Response Error: {:?}", e);
                    (StatusCode::INTERNAL_SERVER_ERROR, e.to_string())
                })?,
            );
        }
    };

    Ok(response)
}

#[allow(unused_variables)]
fn render_lua_error(e: mlua::Error) -> (StatusCode, String) {
    match e {
        mlua::Error::SyntaxError {
            message,
            incomplete_input,
        } => {
            let mut message = message.split("]:");
            message.next();
            (
                StatusCode::INTERNAL_SERVER_ERROR,
                message.next().unwrap_or_default().to_string(),
            )
        }
        mlua::Error::RuntimeError(message) => (StatusCode::INTERNAL_SERVER_ERROR, message),
        mlua::Error::MemoryError(message) => (StatusCode::INTERNAL_SERVER_ERROR, message),
        mlua::Error::SafetyError(message) => (StatusCode::INTERNAL_SERVER_ERROR, message),
        mlua::Error::MemoryControlNotAvailable => (
            StatusCode::INTERNAL_SERVER_ERROR,
            "MemoryControlNotAvailable".to_string(),
        ),
        mlua::Error::RecursiveMutCallback => (
            StatusCode::INTERNAL_SERVER_ERROR,
            "RecursiveMutCallback".to_string(),
        ),
        mlua::Error::CallbackDestructed => (
            StatusCode::INTERNAL_SERVER_ERROR,
            "CallbackDestructed".to_string(),
        ),
        mlua::Error::StackError => (StatusCode::INTERNAL_SERVER_ERROR, "StackError".to_string()),
        mlua::Error::BindError => (StatusCode::INTERNAL_SERVER_ERROR, "BindError".to_string()),
        mlua::Error::BadArgument {
            to,
            pos,
            name,
            cause,
        } => (
            StatusCode::INTERNAL_SERVER_ERROR,
            format!(
                "Bad Argument:\n\tto: {}\n\tpos: {}\n\tname: {}\n\tcause: {:?}",
                to.unwrap_or_default(),
                pos,
                name.unwrap_or_default(),
                cause
            ),
        ),
        mlua::Error::ToLuaConversionError { from, to, message } => (
            StatusCode::INTERNAL_SERVER_ERROR,
            message.unwrap_or_default(),
        ),
        mlua::Error::FromLuaConversionError { from, to, message } => (
            StatusCode::INTERNAL_SERVER_ERROR,
            message.unwrap_or_default(),
        ),
        mlua::Error::CoroutineUnresumable => (StatusCode::INTERNAL_SERVER_ERROR, String::new()),
        mlua::Error::UserDataTypeMismatch => (StatusCode::INTERNAL_SERVER_ERROR, String::new()),
        mlua::Error::UserDataDestructed => (StatusCode::INTERNAL_SERVER_ERROR, String::new()),
        mlua::Error::UserDataBorrowError => (StatusCode::INTERNAL_SERVER_ERROR, String::new()),
        mlua::Error::UserDataBorrowMutError => (StatusCode::INTERNAL_SERVER_ERROR, String::new()),
        mlua::Error::MetaMethodRestricted(message) => (StatusCode::INTERNAL_SERVER_ERROR, message),
        mlua::Error::MetaMethodTypeError {
            method,
            type_name,
            message,
        } => (
            StatusCode::INTERNAL_SERVER_ERROR,
            message.unwrap_or_default(),
        ),
        mlua::Error::MismatchedRegistryKey => (StatusCode::INTERNAL_SERVER_ERROR, String::new()),
        mlua::Error::CallbackError { traceback, cause } => (
            StatusCode::INTERNAL_SERVER_ERROR,
            format!("{}\n\n{:#?}", traceback, cause),
        ),
        mlua::Error::PreviouslyResumedPanic => (StatusCode::INTERNAL_SERVER_ERROR, String::new()),
        mlua::Error::SerializeError(message) => (StatusCode::INTERNAL_SERVER_ERROR, message),
        mlua::Error::DeserializeError(message) => (StatusCode::INTERNAL_SERVER_ERROR, message),
        mlua::Error::ExternalError(error) => {
            (StatusCode::INTERNAL_SERVER_ERROR, format!("{:#?}", error))
        }
        mlua::Error::WithContext { context, cause } => (
            StatusCode::INTERNAL_SERVER_ERROR,
            format!("{}\n\n{:#?}", context, cause),
        ),
        _ => (
            StatusCode::INTERNAL_SERVER_ERROR,
            String::from("Unknown Error"),
        ),
    }
}

async fn handler(request: Request) -> Result<Html<String>, (StatusCode, String)> {
    let start = Instant::now();
    let base_dir = BASE_DIR.get().unwrap();
    let mut path = base_dir.clone();

    let uri = urlencoding::decode(&request.uri().path()[1..]).map_err(|e| {
        println!("{:?}", e);
        (StatusCode::INTERNAL_SERVER_ERROR, e.to_string())
    })?;

    path.push(&*uri);
    let full_path = fs::canonicalize(&path).map_err(|e| {
        println!("{:?}", e);
        (StatusCode::NOT_FOUND, "404: Not Found".to_string())
    })?;

    let full_base_path = fs::canonicalize(base_dir).map_err(|e| {
        println!("{:?}", e);
        (StatusCode::NOT_FOUND, "404: Not Found".to_string())
    })?;

    if !full_path.starts_with(full_base_path) {
        return Err((StatusCode::BAD_REQUEST, "400: Bad Request".to_string()));
    }

    let body = match render_dir(&path, start).await {
        Ok(body) => Ok(body),
        Err(e) => {
            let ioerror: Option<&std::io::Error> = e.downcast_ref();
            if let Some(ioerror) = ioerror {
                if ioerror.kind() == ErrorKind::NotFound {
                    Err((StatusCode::NOT_FOUND, "404: Not Found".to_string()))
                } else {
                    println!("{:?}", ioerror);
                    Err((StatusCode::INTERNAL_SERVER_ERROR, e.to_string()))
                }
            } else {
                println!("{:?}", e);
                Err((StatusCode::INTERNAL_SERVER_ERROR, e.to_string()))
            }
        }
    };

    Ok(Html(body?))
}

async fn render_dir(path: &PathBuf, time: Instant) -> AResult<String> {
    let base_dir = BASE_DIR.get().unwrap();
    let time2 = Instant::now();
    let base = base_dir.display().to_string();
    let dirname = path
        .display()
        .to_string()
        .replace(&base, "")
        .replace('\\', "/");

    let filename = path.file_name().unwrap().to_string_lossy().to_string();
    let description = PathBuf::from_str(&format!(
        "{}/.{}.info",
        path.parent().unwrap().to_string_lossy(),
        filename
    ))?;
    let description = if description.exists() {
        Some(fs::read_to_string(description)?)
    } else {
        None
    };

    let directories = get_directories(path).await?;
    let files = get_files(path).await?;

    let tera = &*TERA;

    let mut context = Context::new();
    context.insert("description", &description);
    context.insert("dirname", &dirname);
    context.insert("directories", &directories);
    context.insert("files", &files);

    let loading = time2.elapsed() + time.elapsed();
    context.insert("loading", &loading.as_millis());
    Ok(tera.render("index.html.jinja", &context)?)
}

async fn get_directories(path: &PathBuf) -> AResult<Vec<FileInfo>> {
    let mut contents = ReadDirStream::new(tokio::fs::read_dir(&path).await?);

    let mut files = vec![];

    while let Some(file) = contents.next().await {
        let file = file?;
        let metadata = file.metadata().await?;
        if !metadata.is_dir() {
            continue;
        }
        let filename = file.file_name().to_string_lossy().to_string();
        if filename.starts_with('.') {
            continue;
        }

        // let created: OffsetDateTime = metadata.created()?.into();
        let modified: OffsetDateTime = metadata.modified()?.into();

        let description = PathBuf::from_str(&format!(
            "{}/.{}.info",
            file.path().parent().unwrap().to_string_lossy(),
            filename
        ))?;

        let description = if description.exists() {
            Some(fs::read_to_string(description)?)
        } else {
            None
        };

        files.push(FileInfo {
            filename,
            format: "DIR".to_string(),
            size: 0,
            icon: None,
            // created,
            modified,
            description,
        });
    }

    Ok(files)
}

async fn get_files(path: &PathBuf) -> AResult<Vec<FileInfo>> {
    let mut contents = ReadDirStream::new(tokio::fs::read_dir(&path).await?);

    let mut files = vec![];

    while let Some(file) = contents.next().await {
        let file = file?;
        let metadata = file.metadata().await?;
        if metadata.is_dir() {
            continue;
        }
        let filename = file.file_name().to_string_lossy().to_string();
        if filename.starts_with('.') {
            continue;
        }
        let format = file_format::FileFormat::from_file(file.path())?;
        let format = format.media_type();

        let icon = if format == "application/x-dosexec" {
            // ..
            extract_icon(&file.path()).ok().flatten()
        } else {
            None
        };

        // let created: OffsetDateTime = metadata.created()?.into();
        let modified: OffsetDateTime = metadata.modified()?.into();

        let description = PathBuf::from_str(&format!(
            "{}/.{}.info",
            file.path().parent().unwrap().to_string_lossy(),
            filename
        ))?;

        let description = if description.exists() {
            Some(fs::read_to_string(description)?)
        } else {
            None
        };

        files.push(FileInfo {
            filename,
            format: format.to_string(),
            size: metadata.len() as usize,
            icon,
            // created,
            modified,
            description,
        });
    }

    Ok(files)
}
